from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

from kleur import Color, Colored
from kleur.interpol import LinearMapping, mapped

from .utils import ArgsParser, check_integer_in_range, print_lines

if TYPE_CHECKING:
    from argparse import Namespace
    from collections.abc import Iterator


def _css_color_comment(color: Color) -> Colored:
    return Colored(f"#{color.as_hex} --> {color}", color, color.contrasting_shade)


class LinesGenerator(ABC):
    def __init__(self, args: Namespace) -> None:
        self._include_input = args.include_input_shades
        self._include_bw = args.include_black_and_white
        self._label = args.label
        steps = args.number_of_shades + 1
        shades_range = (0, steps + 1) if self._include_bw else (1, steps)
        self._shades = [s / steps for s in range(*shades_range)]
        self._generated_shades: set[int] = set()

    @abstractmethod
    def _comment_lines(self) -> Iterator[str]: ...

    @abstractmethod
    def _colors(self) -> Iterator[tuple[Color, str]]: ...

    def lines(self) -> Iterator[str]:
        yield "/*"
        yield from self._comment_lines()
        yield "*/"

        colors = {f"{c.lightness * 100:.0f}": (c, i) for c, i in self._colors()}
        sorted_colors = sorted(colors.items())
        n_max, _ = sorted_colors[-1]
        n_digits = len(str(n_max))
        for n, (color, input_indication) in sorted_colors:
            css_var = f"--{self._label}-{n.zfill(n_digits)}: #{color.as_hex};"
            comment = f"--> {color}"
            if input_indication:
                comment += f" <-- {input_indication}"
            k = color.contrasting_shade
            fg, bg = (color, k) if input_indication else (k, color)
            yield str(Colored(f"{css_var} /* {comment} */", fg, bg))


class LinesGeneratorOneColor(LinesGenerator):
    def __init__(self, args: Namespace) -> None:
        super().__init__(args)
        self._input = Color.from_hex(args.color1)

    def _comment_lines(self) -> Iterator[str]:
        yield f"Based on: {_css_color_comment(self._input)}"

    def _colors(self) -> Iterator[tuple[Color, str]]:
        """Generate shades of a color."""
        for s in self._shades:
            yield self._input.shade(s), ""
        if self._include_input:
            yield self._input, "input"


def normalize_color(c1: Color, c2: Color) -> Color:
    return c1 if round(c1.saturation, 2) else Color(c2.hue, c1.saturation, c1.lightness)


class LinesGeneratorTwoColors(LinesGenerator):
    def __init__(self, args: Namespace) -> None:
        super().__init__(args)
        self._dynamic_range = args.dynamic_range / 100
        c1, c2 = Color.from_hex(args.color1), Color.from_hex(args.color2)
        c1 = normalize_color(c1, c2)
        c2 = normalize_color(c2, c1)
        self._dark_input, self._bright_input = sorted((c1, c2))

    def _comment_lines(self) -> Iterator[str]:
        yield "Based on:"
        yield f"- Darkest:   {_css_color_comment(self._dark_input)}"
        yield f"- Brightest: {_css_color_comment(self._bright_input)}"

    def _colors(self) -> Iterator[tuple[Color, str]]:
        """
        Generate shades based on two colors.

        The dynamic range specifies to what degree the hue
        of the input colors will be used as boundaries:
        - dynamic range 0 (0%):
            The shades will interpolate (or extrapolate) between the input colors
        - dynamic range between 0 and 1 (between 0% and 100%):
            The shades will interpolate (or extrapolate) between
            darker / brighter shades of the input colors
        - dynamic range 1 (100%):
            The shades will interpolate (or extrapolate) between
            the darkest & brightest shades of the input colors
        """
        old_dark_shade = self._dark_input.lightness
        new_dark_shade = mapped(self._dynamic_range, (old_dark_shade, 0))
        new_dark = self._dark_input.shade(new_dark_shade)

        old_bright_shade = self._bright_input.lightness
        new_bright_shade = mapped(self._dynamic_range, (old_bright_shade, 1))
        new_bright = self._bright_input.shade(new_bright_shade)

        shade_mapping = LinearMapping(new_dark_shade, new_bright_shade)

        def blend(lightness: float) -> Color:
            return new_dark.blend(new_bright, shade_mapping.position_of(lightness))

        for s in self._shades:
            yield blend(s), ""

        if self._include_input:
            i = "same {} as {} input"

            yield blend(old_dark_shade), i.format("shade", "darkest")
            if old_dark_shade != new_dark_shade:
                yield new_dark, i.format("hue", "darkest")

            yield blend(old_bright_shade), i.format("shade", "brightest")
            if old_bright_shade != new_bright_shade:
                yield new_bright, i.format("hue", "brightest")


class CssArgsParser(ArgsParser):
    name = "css"

    def _parse_args(self) -> None:
        self._parser.add_argument("-l", "--label", type=str, default="color")
        self._parser.add_argument("-c", "--color1", type=str, required=True)
        self._parser.add_argument("-k", "--color2", type=str)
        self._parser.add_argument(
            "-s", "--number-of-shades", type=check_integer_in_range(1, 99), default=19
        )
        self._parser.add_argument(
            "-b", "--include-black-and-white", action="store_true", default=False
        )
        self._parser.add_argument(
            "-i", "--include-input-shades", action="store_true", default=False
        )
        self._parser.add_argument(
            "-d", "--dynamic-range", type=check_integer_in_range(0, 100), default=0
        )

    def _run_command(self, args: Namespace) -> None:
        gen_cls = LinesGeneratorTwoColors if args.color2 else LinesGeneratorOneColor
        print_lines(gen_cls(args).lines())



from typing import TYPE_CHECKING

from kleur import AltColors, Color, Colored, Colors

from .utils import (
    ArgsParser,
    check_integer_in_range,
    get_class_vars,
    parse_key_value_pair,
    print_lines,
    try_convert,
)

if TYPE_CHECKING:
    from argparse import Namespace
    from collections.abc import Iterable, Iterator

BLOCK_WIDTH = 8


class LinesGenerator:
    def __init__(self, args: Namespace) -> None:
        n_shades, n_vibrances = args.number_of_shades, args.number_of_vibrances
        self._shades = [s / (n_shades + 1) for s in range(1, n_shades + 1)]
        self._vibrances = [v / n_vibrances for v in range(1, n_vibrances + 1)]

        hues: dict[int, str] = {}

        if args.merge_with_default_theme or not args.colors:
            theme_cls = AltColors if args.alt_default_theme else Colors
            for name, color in get_class_vars(theme_cls, Color).items():
                hues[round(color.hue * 360)] = name

        for name, hue in args.colors:
            hues[try_convert(int, hue, default=333)] = name

        self._hues = sorted(hues.items())

    def _blocks(
        self, name: str, vibrance: float, hue: float = None
    ) -> Iterable[Colored]:
        color = Color(hue or 0, vibrance)

        for s in self._shades:
            c = color.shade(s)
            yield Colored(c.as_hex.center(BLOCK_WIDTH), c.contrasting_shade, c)

        hue_str = "   " if hue is None else f"{hue * 360:03.0f}"
        yield Colored(f" {hue_str} {name}", color)

    def _line(self, name: str, vibrance: float, hue: float = None) -> str:
        return "".join(str(c) for c in self._blocks(name, vibrance, hue))

    def lines(self) -> Iterator[str]:
        yield self._line("grey", 0)

        for vibrance in self._vibrances:
            yield "".join(f"{s:.2%}".center(BLOCK_WIDTH) for s in self._shades)

            for hue, name in self._hues:
                yield self._line(name, vibrance, hue / 360)


class ThemeArgsParser(ArgsParser):
    name = "theme"

    def _parse_args(self) -> None:
        self._parser.add_argument(
            "-c",
            "--colors",
            nargs="+",
            metavar="NAME=HUE (1-360)",
            type=parse_key_value_pair,
            default={},
        )
        self._parser.add_argument(
            "-m", "--merge-with-default-theme", action="store_true", default=False
        )
        self._parser.add_argument(
            "-a", "--alt-default-theme", action="store_true", default=False
        )
        self._parser.add_argument(
            "-s", "--number-of-shades", type=check_integer_in_range(1, 99), default=9
        )
        self._parser.add_argument(
            "-v", "--number-of-vibrances", type=check_integer_in_range(1, 99), default=2
        )

    def _run_command(self, args: Namespace) -> None:
        print_lines(LinesGenerator(args).lines())



from dataclasses import astuple, dataclass, replace
from functools import cached_property, total_ordering
from typing import TYPE_CHECKING

from hsluv import hex_to_hsluv, hsluv_to_hex, hsluv_to_rgb, rgb_to_hsluv

from .interpol import mapped, mapped_cyclic, trim, trim_cyclic

if TYPE_CHECKING:
    from collections.abc import Callable

_INCREASE_FACTOR = 2**0.5


def _normalize_rgb_hex(rgb_hex: str) -> str:
    """
    Try to normalize a hex string into a rrggbb hex.

    :param rgb_hex: RGB hex string (may start with '#')
    :return: rrggbb hex

    >>> _normalize_rgb_hex("3")
    '333333'
    >>> _normalize_rgb_hex("03")
    '030303'
    >>> _normalize_rgb_hex("303")
    '330033'
    >>> _normalize_rgb_hex("808303")
    '808303'
    """
    rgb_hex, r, g, b = rgb_hex.removeprefix("#").lower(), "", "", ""

    match len(rgb_hex):
        case 1:
            # 3 -> r=33, g=33, b=33
            r = g = b = rgb_hex * 2

        case 2:
            # 03 -> r=03, g=03, b=03
            r = g = b = rgb_hex

        case 3:
            # 303 -> r=33, g=00, b=33
            r1, g1, b1 = iter(rgb_hex)
            r, g, b = r1 * 2, g1 * 2, b1 * 2

        case 6:
            # 808303 -> r=80, g=83, b=03
            r1, r2, g1, g2, b1, b2 = iter(rgb_hex)
            r, g, b = r1 + r2, g1 + g2, b1 + b2

        case _:
            raise ValueError(rgb_hex)

    return f"{r}{g}{b}"


type RGB = tuple[int, int, int]


@dataclass(frozen=True)
class _HSLuv:
    hue: float
    saturation: float
    lightness: float

    def __repr__(self) -> str:
        return f"HSLuv({self.hue:.2f}°, {self.saturation:.2f}%, {self.lightness:.2f}%)"

    @classmethod
    def from_hex(cls, rgb_hex: str) -> _HSLuv:
        return cls(*hex_to_hsluv(f"#{rgb_hex}"))

    @cached_property
    def as_hex(self) -> str:
        return hsluv_to_hex(astuple(self))[1:]

    @classmethod
    def from_rgb(cls, rgb: RGB) -> _HSLuv:
        r, g, b = rgb
        return cls(*rgb_to_hsluv((r / 255, g / 255, b / 255)))

    @cached_property
    def as_rgb(self) -> RGB:
        r, g, b = hsluv_to_rgb(astuple(self))
        return round(r * 255), round(g * 255), round(b * 255)


@total_ordering
@dataclass(frozen=True)
class Color:
    hue: float = 0  # 0 - 1 (full circle angle)
    saturation: float = 1  # 0 - 1 (ratio)
    lightness: float = 0.5  # 0 - 1 (ratio)

    def __post_init__(self) -> None:
        self._normalize_values()

    def _normalize_values(self) -> None:
        """
        Make sure all color values are in a valid range.

        object.__setattr__() is one of the awkward options (*) we have,
        when we want to set attributes in a frozen dataclass (which will raise a
        FrozenInstanceError when its own __setattr__() or __delattr__() is invoked).

        *) Another option could be to move the attributes to a super class and
        call super().__init__() here.
        """
        funcs: dict[str, Callable[[float], float]] = {
            "hue": trim_cyclic,
            "saturation": trim,
            "lightness": trim,
        }
        for name, func in funcs.items():
            object.__setattr__(self, name, func(getattr(self, name)))

    def __repr__(self) -> str:
        return repr(self._as_hsluv)

    def __lt__(self, other: Color) -> bool:
        return self.as_sortable_tuple < other.as_sortable_tuple

    @cached_property
    def as_sortable_tuple(self) -> tuple[float, float, float]:
        """Will decide the sort order."""
        return self.lightness, self.saturation, self.hue

    @classmethod
    def _from_hsluv(cls, hsluv: _HSLuv) -> Color:
        return cls(hsluv.hue / 360, hsluv.saturation / 100, hsluv.lightness / 100)

    @cached_property
    def _as_hsluv(self) -> _HSLuv:
        return _HSLuv(self.hue * 360, self.saturation * 100, self.lightness * 100)

    @classmethod
    def from_hex(cls, rgb_hex: str) -> Color:
        """
        Create a Color from an RGB hex string.

        :param rgb_hex: RGB hex string (may start with '#')
        :return: Color instance

        >>> c1 = Color.from_hex("808303")
        >>> c1.as_hex
        '808303'
        >>> c1.as_rgb
        (128, 131, 3)
        >>> c2 = Color.from_hex("0af")
        >>> c2.as_hex
        '00aaff'
        >>> c2.as_rgb
        (0, 170, 255)
        """
        return cls._from_hsluv(_HSLuv.from_hex(_normalize_rgb_hex(rgb_hex)))

    @cached_property
    def as_hex(self) -> str:
        return self._as_hsluv.as_hex

    @classmethod
    def from_rgb(cls, rgb: RGB) -> Color:
        """
        Create a Color from RGB values.

        :param rgb: RGB instance
        :return: Color instance

        >>> c1 = Color.from_rgb((128, 131, 3))
        >>> c1.as_hex
        '808303'
        >>> c1.as_rgb
        (128, 131, 3)
        >>> c2 = Color.from_rgb((0, 170, 255))
        >>> c2.as_hex
        '00aaff'
        >>> c2.as_rgb
        (0, 170, 255)
        """
        return cls._from_hsluv(_HSLuv.from_rgb(rgb))

    @cached_property
    def as_rgb(self) -> RGB:
        return self._as_hsluv.as_rgb

    def adjust(
        self, *, hue: float = None, saturation: float = None, lightness: float = None
    ) -> Color:
        return Color(
            self.hue + (hue or 0),
            self.saturation * (saturation or 1),
            self.lightness * (lightness or 1),
        )

    def shade(self, lightness: float) -> Color:
        return replace(self, lightness=lightness)

    def saturated(self, saturation: float) -> Color:
        return replace(self, saturation=saturation)

    @cached_property
    def very_bright(self) -> Color:
        return self.shade(5 / 6)

    @cached_property
    def bright(self) -> Color:
        return self.shade(4 / 6)

    @cached_property
    def dark(self) -> Color:
        return self.shade(2 / 6)

    @cached_property
    def very_dark(self) -> Color:
        return self.shade(1 / 6)

    def brighter(self, relative_amount: float = _INCREASE_FACTOR) -> Color:
        return self.adjust(lightness=relative_amount)

    @cached_property
    def slightly_brighter(self) -> Color:
        return self.brighter(_INCREASE_FACTOR**0.5)

    @cached_property
    def much_brighter(self) -> Color:
        return self.brighter(_INCREASE_FACTOR**2)

    def darker(self, relative_amount: float = _INCREASE_FACTOR) -> Color:
        return self.brighter(1 / relative_amount)

    @cached_property
    def slightly_darker(self) -> Color:
        return self.darker(_INCREASE_FACTOR**0.5)

    @cached_property
    def much_darker(self) -> Color:
        return self.darker(_INCREASE_FACTOR**2)

    def blend(self, other: Color, amount: float = 0.5) -> Color:
        return Color(
            mapped_cyclic(amount, (self.hue, other.hue), period=1),
            mapped(amount, (self.saturation, other.saturation)),
            mapped(amount, (self.lightness, other.lightness)),
        )

    @cached_property
    def contrasting_shade(self) -> Color:
        """
        Color with a lightness that contrasts with the current color.

        Color with a 50% lower or higher lightness than the current color,
        while maintaining the same hue and saturation (so it can for example
        be used as background color).

        :return: Color representation of the contrasting shade

        >>> Color.from_hex("08f").contrasting_shade.as_hex
        '001531'
        >>> Color.from_hex("0f8").contrasting_shade.as_hex
        '006935'
        >>> Color.from_hex("80f").contrasting_shade.as_hex
        'ebe4ff'
        >>> Color.from_hex("8f0").contrasting_shade.as_hex
        '366b00'
        >>> Color.from_hex("f08").contrasting_shade.as_hex
        '2b0012'
        >>> Color.from_hex("f80").contrasting_shade.as_hex
        '4a2300'
        """
        return self.shade((self.lightness + 0.5) % 1)

    @cached_property
    def contrasting_hue(self) -> Color:
        """
        Color with a hue that contrasts with the current color.

        Color with a 180° different hue than the current color,
        while maintaining the same saturation and perceived lightness.

        :return: Color representation of the contrasting hue

        >>> Color.from_hex("08f").contrasting_hue.as_hex
        '9c8900'
        >>> Color.from_hex("0f8").contrasting_hue.as_hex
        'ffd1f5'
        >>> Color.from_hex("80f").contrasting_hue.as_hex
        '5c6900'
        >>> Color.from_hex("8f0").contrasting_hue.as_hex
        'f6d9ff'
        >>> Color.from_hex("f08").contrasting_hue.as_hex
        '009583'
        >>> Color.from_hex("f80").contrasting_hue.as_hex
        '00b8d1'
        """
        return self.adjust(hue=0.5)


def c(h: int) -> Color:
    return Color(h / 360)


@dataclass(frozen=True)
class ColorTheme:
    """
    Helper class to make a Colors class return Color objects.

    Can be overridden by specifying a custom anum with
    a different set of hues (in degrees):
    >>> class MyColors(ColorTheme):
    ...     tomato = c(15)
    ...     turquoise = c(175)
    >>> MyColors.tomato
    HSLuv(15.00°, 100.00%, 50.00%)
    >>> MyColors.grey
    HSLuv(0.00°, 0.00%, 50.00%)
    """

    grey = Color(saturation=0)
    black = grey.shade(0)
    white = grey.shade(1)


class Colors(ColorTheme):
    """Highly opinionated (though carefully selected) color theme."""

    red = c(12)
    orange = c(33)
    yellow = c(69)
    poison = c(101)
    green = c(127)
    ocean = c(190)
    blue = c(248)
    indigo = c(267)
    purple = c(281)
    pink = c(329)

    brown = orange.blend(yellow, 0.25).saturated(0.5)


class AltColors(ColorTheme):
    """Alternative color theme."""

    red = c(10)
    orange = c(35)
    yellow = c(75)
    poison = c(100)
    green = c(126)
    ocean = c(184)
    blue = c(242)
    indigo = c(268)
    purple = c(280)
    pink = c(325)

    brown = orange.blend(yellow, 0.3).saturated(0.5)
